// print isosurface table statistics

#include <fstream>
#include <iostream>
#include <stdio.h> 
#include <sstream>

#include "isotable.h"

using namespace ISOTABLE;
using namespace std;

typedef float COORD_TYPE;

// global variables
const char * isotable_filename;

// routines
void parse_command_line(int argc, char **argv);
void usage_error();

int main(int argc, char **argv)
{
  ISOSURFACE_TABLE isotable(3);

  parse_command_line(argc, argv);

  try {
    ifstream isotable_file(isotable_filename, ios::in);

    if (!isotable_file) {
      cerr << "Unable to read isosurface table file "
	   << isotable_filename << "." << endl;
      exit(30);
    };

    isotable.ReadText(isotable_file);
  }
  catch (ISOSURFACE_TABLE_ERROR error) {
    cerr << "Error reading isosurface table from " 
	 << isotable_filename << "." << endl;
    cerr << error.Msg() << endl;
    cerr << "Exiting." << endl;
    exit(35);
  };

  long max_s = 0;
  long total_s = 0;
  for (long i = 0; i < isotable.NumTableEntries(); i++) {
    long n = isotable.NumSimplices(i);
    if (max_s < n) max_s = n;
    total_s += n;
  };
  double avg_s = ((double) total_s)/((double) isotable.NumTableEntries());

  cout << "Dimension = " << isotable.Dimension() << endl;
  cout << "Polyhedron:" << endl;
  cout << "  # Vertices = " << isotable.Polyhedron().NumVertices() << endl;
  cout << "  # Edges = " << isotable.Polyhedron().NumEdges() << endl;
  cout << "  # Facets = " << isotable.Polyhedron().NumFacets() << endl;
  cout << "Table:" << endl;
  cout << "  # Entries = " << isotable.NumTableEntries() << endl;
  cout << "  Max # simplices per entry = " << max_s << endl;
  cout << "  Avg # simplices per entry = " << avg_s << endl;
}

void parse_command_line(int argc, char **argv)
{
  if (argc != 2) usage_error();
  isotable_filename = argv[1];
}

void usage_error()
{
  cerr << "Usage: isotablestat {isotable file}" 
       << endl;
  exit(10);
}
